#include "nrf_drv_spi.h"
#include "nrf_drv_timer.h"
#include "nrf_drv_ppi.h"
#include "nrf_drv_gpiote.h"
#include "app_util_platform.h"
#include "nrf_gpio.h"
#include "nrf_delay.h"
#include "boards.h"
#include "app_error.h"
#include <string.h>
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"

#define CNVST_GPIOTE_PIN 31 // conversion start pin on ADC
#define NUM_SAMPLES      5  // number of samples to read in a window

#define SPI_INSTANCE0  0 /**< SPI instance index. */
static const nrf_drv_spi_t spi0 = NRF_DRV_SPI_INSTANCE(SPI_INSTANCE0);  /**< SPI instance. */

static volatile bool spi_xfer_done;  /**< Flag used to indicate that SPI instance completed the transfer. */

static nrf_drv_timer_t timer_adc     = NRF_DRV_TIMER_INSTANCE(1); // timer for adc sampling rate
static nrf_drv_timer_t timer_counter = NRF_DRV_TIMER_INSTANCE(2); // timer for counting samples read

// Variable holding PPI channel number
nrf_ppi_channel_t ppi_channel_spi_start;
nrf_ppi_channel_t ppi_channel_spi_end;
nrf_ppi_channel_t ppi_channel_spi_count;

#define TEST_STRING "Hi"
static uint8_t       m_tx_buf[] = TEST_STRING;           /**< TX buffer. */
static uint8_t       m_rx_buf[2];    /**< RX buffer. */
static const uint8_t m_length = sizeof(m_rx_buf);        /**< Transfer length. */

// Define a type with a two dimensional array for easyDMA to use for SPI rx
typedef struct ArrayList
{
    uint8_t buffer[2];
}array_list_t;

array_list_t rx_buf_array[NUM_SAMPLES];

uint16_t count = 0; // count how many samples have been read
uint16_t samples[NUM_SAMPLES];
float    samples_volts[NUM_SAMPLES];

const float LSB = 5.0/65536.0; // LSB for ADC 16bit sample to voltage conversion

bool array_list_mode = true;

void spi_sampling_enable(void)
{
    ret_code_t err_code;

    // Enable sampling SPI and toggling CNVST pin
    nrf_drv_gpiote_out_task_enable(CNVST_GPIOTE_PIN);
    err_code = nrf_drv_ppi_channel_enable(ppi_channel_spi_start);
    err_code = nrf_drv_ppi_channel_enable(ppi_channel_spi_end);
    err_code = nrf_drv_ppi_channel_enable(ppi_channel_spi_count);
    nrf_drv_timer_enable(&timer_adc);
    if (array_list_mode)
    {
        nrf_drv_timer_enable(&timer_counter);
    }
}

void spi_sampling_disable(void)
{
    ret_code_t err_code;
    
    // Disable sampling SPI and toggling CNVST pin
    nrf_drv_gpiote_out_task_disable(CNVST_GPIOTE_PIN);
    err_code = nrf_drv_ppi_channel_disable(ppi_channel_spi_start);
    err_code = nrf_drv_ppi_channel_disable(ppi_channel_spi_end);
    err_code = nrf_drv_ppi_channel_disable(ppi_channel_spi_count);
    nrf_drv_timer_disable(&timer_adc);
    if (array_list_mode)
    {
        nrf_drv_timer_disable(&timer_counter);
    }
}

/**
 * @brief SPI user event handler.
 * @param event
 */
void spi_event_handler(nrf_drv_spi_evt_t const * p_event,
                       void *                    p_context)
{
    if(count==NUM_SAMPLES)
    {
        spi_sampling_disable();
        spi_xfer_done = true;
    }
    else
    {
        samples[count] = ((uint16_t)(m_rx_buf[0]<<8) | m_rx_buf[1]);
        count++;
    }
}

// SPI transfer start handler for timer_adc. Not used
void timer_adc_dummy_handler(nrf_timer_event_t event_type, void * p_context)
{ // Do nothing
}

void timer_count_handler(nrf_timer_event_t event_type, void * p_context)
{
    spi_sampling_disable();
    spi_xfer_done = true;
}

void spi_sampling_init(void)
{
    printf("a\n");

    ret_code_t err_code;

    nrf_drv_gpiote_out_config_t gpiote_config = GPIOTE_CONFIG_OUT_TASK_TOGGLE(true);  
    err_code = nrf_drv_gpiote_out_init(CNVST_GPIOTE_PIN, &gpiote_config);
    
    // Initialise rx buffer as all 0
    memset(m_rx_buf, 0, m_length);
    memset(rx_buf_array, 0, NUM_SAMPLES*2);

    nrf_drv_spi_config_t spi_config = NRF_DRV_SPI_DEFAULT_CONFIG;
    spi_config.mode     = NRF_DRV_SPI_MODE_1;
    spi_config.frequency = NRF_DRV_SPI_FREQ_8M;
    spi_config.ss_pin   = NRF_DRV_SPI_PIN_NOT_USED;
    spi_config.miso_pin = 30; //SPI_MISO_PIN;
    spi_config.mosi_pin = NRF_DRV_SPI_PIN_NOT_USED;
    spi_config.sck_pin  = 28; //SPI_SCK_PIN;

    // Set up next transfer sequence
    if (array_list_mode)
    {
        APP_ERROR_CHECK(nrf_drv_spi_init(&spi0, &spi_config, NULL, NULL));
        nrf_drv_spi_xfer_desc_t xfer = NRF_DRV_SPI_XFER_TRX(m_tx_buf, m_length, (uint8_t*)rx_buf_array, m_length);
        
        printf("b\n");
        // Set flags for next transfer sequence
         uint32_t flags = NRF_DRV_SPI_FLAG_HOLD_XFER     |
                          NRF_DRV_SPI_FLAG_REPEATED_XFER |
                          NRF_DRV_SPI_FLAG_RX_POSTINC    |
                          NRF_DRV_SPI_FLAG_NO_XFER_EVT_HANDLER;
        printf("c\n");

        // Pass configurations and flags to SPIM driver.
        // The SPIM is now on and prepared, but not actually started.
        err_code = nrf_drv_spi_xfer(&spi0, &xfer, flags);
        
        APP_ERROR_CHECK(err_code);
    }
    else
    {
        APP_ERROR_CHECK(nrf_drv_spi_init(&spi0, &spi_config, spi_event_handler, NULL));
        nrf_drv_spi_xfer_desc_t xfer = NRF_DRV_SPI_XFER_TRX(m_tx_buf, m_length, m_rx_buf, m_length);
        
        // Set flags for next transfer sequence
        uint32_t flags = NRF_DRV_SPI_FLAG_HOLD_XFER |
                     NRF_DRV_SPI_FLAG_REPEATED_XFER;
    
        // Pass configurations and flags to SPIM driver.
        // The SPIM is now on and prepared, but not actually started.
        err_code = nrf_drv_spi_xfer(&spi0, &xfer, flags);
        APP_ERROR_CHECK(err_code);
    }

    nrf_drv_timer_config_t timer_adc_cfg = NRF_DRV_TIMER_DEFAULT_CONFIG;
    timer_adc_cfg.frequency = NRF_TIMER_FREQ_16MHz;
    // Initializing the timer triggering the SPI transfers. 
    err_code = nrf_drv_timer_init(&timer_adc, &timer_adc_cfg, timer_adc_dummy_handler);
    APP_ERROR_CHECK(err_code);
    
    // Sample every 10 us (100 kHz)
    // uint32_t adc_sample_period = nrf_drv_timer_us_to_ticks(&timer_instance_twi_start, 10); 
    uint32_t adc_sample_period = 159;

    // Configuring the timer triggering the TWI transfers
    nrf_drv_timer_extended_compare(&timer_adc, NRF_TIMER_CC_CHANNEL0, adc_sample_period, NRF_TIMER_SHORT_COMPARE0_CLEAR_MASK, false);
    
    uint32_t gpiote_task_addr = nrf_drv_gpiote_out_task_addr_get(CNVST_GPIOTE_PIN);    
    uint32_t timer_start_compare_event_addr = nrf_drv_timer_compare_event_address_get(&timer_adc, NRF_TIMER_CC_CHANNEL0);
    uint32_t spi_start_task_addr   = nrf_drv_spi_start_task_get(&spi0);
    uint32_t spi_end_evt_addr = nrf_drv_spi_end_event_get(&spi0);
    
    // Allocate PPI channel to toggle on CNVST pin on timer_adc events (100kHz)
    err_code = nrf_drv_ppi_channel_alloc(&ppi_channel_spi_start);
    APP_ERROR_CHECK(err_code);
    err_code = nrf_drv_ppi_channel_assign(ppi_channel_spi_start, timer_start_compare_event_addr, gpiote_task_addr);
    APP_ERROR_CHECK(err_code);

    // Use same PPI channel to start SPI task on same timer_adc event
    err_code = nrf_drv_ppi_channel_fork_assign(ppi_channel_spi_start, spi_start_task_addr);
    APP_ERROR_CHECK(err_code);
    				
    // Allocate another PPI channel to toggle off CNVST pin at SPI end event                     															 																		 
    err_code = nrf_drv_ppi_channel_alloc(&ppi_channel_spi_end);
    APP_ERROR_CHECK(err_code);
    err_code = nrf_drv_ppi_channel_assign(ppi_channel_spi_end, spi_end_evt_addr, gpiote_task_addr);
    APP_ERROR_CHECK(err_code);

    if (array_list_mode)
    {
        // set up another timer in counter mode for counting SPI transfers
        nrf_drv_timer_config_t counter_config = NRF_DRV_TIMER_DEFAULT_CONFIG;
        counter_config.mode = NRF_TIMER_MODE_COUNTER;
        // Initializing the SPI counter. 
        err_code = nrf_drv_timer_init(&timer_counter, &counter_config, timer_count_handler);
        APP_ERROR_CHECK(err_code);
        
        nrf_drv_timer_extended_compare(&timer_counter, NRF_TIMER_CC_CHANNEL0, NUM_SAMPLES, NRF_TIMER_SHORT_COMPARE0_CLEAR_MASK, true);
        
        uint32_t task_address_counter_inc = nrf_drv_timer_task_address_get(&timer_counter, NRF_TIMER_TASK_COUNT);

        // Add another PPI channel for counting SPI transfers - increment counter on timer_adc events
        err_code = nrf_drv_ppi_channel_alloc(&ppi_channel_spi_count);
        APP_ERROR_CHECK(err_code);

        err_code = nrf_drv_ppi_channel_assign(ppi_channel_spi_count, spi_end_evt_addr, task_address_counter_inc);
        APP_ERROR_CHECK(err_code);
    }
}

/**
 * @brief ADC conversion to voltage.
 * @param ADC value
 */
float convertToVoltage(uint16_t ADC)
{
  uint16_t df;
  float xdoffv;
  df = (1<<15);
  
  if(ADC & df)
  {
        uint16_t temp1, temp2;
        temp1 = ~(ADC) + (uint16_t)1;       
        xdoffv = -1 * (temp1 * LSB);        
  }
  else
  {
        xdoffv = (float)ADC * LSB;
  }
  return xdoffv;
}


int main(void)
{
    ret_code_t err_code;
    uint8_t timer_count_value;

    APP_ERROR_CHECK(NRF_LOG_INIT(NULL));
    NRF_LOG_DEFAULT_BACKENDS_INIT();

    err_code = nrf_drv_ppi_init();
    APP_ERROR_CHECK(err_code);

    err_code = nrf_drv_gpiote_init();
    APP_ERROR_CHECK(err_code);

    printf("start\n");
    //nrf_delay_ms(100);

    spi_sampling_init();

    spi_sampling_enable();

    printf("set up done\n");

    while (1)
    {
        /*
        // check if timer counter is ever incrementing
        NRF_TIMER2->TASKS_CAPTURE[0] = 1; 
        timer_count_value = NRF_TIMER2->CC[0];
        NRF_TIMER2->TASKS_CAPTURE[0] = 0;
        printf("count = %d\n", timer_count_value);
        nrf_delay_ms(100);
        */

        if(spi_xfer_done == true)
        {
            for(int i = 0; i < NUM_SAMPLES; i++)
            {
                //printf("%x %x\n", rx_buf_array[i].buffer[0], rx_buf_array[i].buffer[1]);
                //samples[i] = ((uint16_t)(rx_buf_array[i].buffer[0]<<8) | rx_buf_array[i].buffer[1]);
                samples_volts[i] = convertToVoltage(samples[i]);
                printf("%f\n", samples_volts[i]);
            }
            nrf_delay_ms(5);
            
            // reset sampling

            if (array_list_mode)
            {
                // Set up next transfer sequence
                nrf_drv_spi_xfer_desc_t xfer = NRF_DRV_SPI_XFER_TRX(m_tx_buf, m_length, (uint8_t*)rx_buf_array, m_length);
   
                // Set flags for next transfer sequence
                uint32_t flags = NRF_DRV_SPI_FLAG_HOLD_XFER           |
                                 NRF_DRV_SPI_FLAG_RX_POSTINC          |
                                 NRF_DRV_SPI_FLAG_NO_XFER_EVT_HANDLER |
                                 NRF_DRV_SPI_FLAG_REPEATED_XFER;
    
                // Pass configurations and flags to SPIM driver.
                // The SPIM is now on and prepared, but not actually started.
                err_code = nrf_drv_spi_xfer(&spi0, &xfer, flags);
                APP_ERROR_CHECK(err_code);
            }

            count = 0;
            spi_xfer_done = false;
            spi_sampling_enable();
            printf("\n\n");
        }
    }
}
